
from .commons import run_client, get_basic_arg_parser


def _runner(client, path, args):

    services = client.discover_services(uri=path)

    for s in services:
        if args.as_xml:
            print(s.raw.to_xml(pretty_print=True).decode('utf-8'))
        else:
            print(s.raw.to_text())


def discover_services():
    run_client(get_basic_arg_parser(), _runner)



from libtaxii import messages_11 as tm11
from libtaxii import messages_10 as tm10

STREAM_MARKER = 'STREAM'

TAXII_11_NS = 'http://taxii.mitre.org/messages/taxii_xml_binding-1.1'
TAXII_10_NS = 'http://taxii.mitre.org/messages/taxii_xml_binding-1'

XML_11_BINDING = 'urn:taxii.mitre.org:message:xml:1.1'
XML_10_BINDING = 'urn:taxii.mitre.org:message:xml:1.0'
CERT_EU_JSON_10_BINDING = 'urn:cert.europa.eu:message:json:1.0'

VERSIONS = {
    TAXII_11_NS: XML_11_BINDING,
    TAXII_10_NS: XML_10_BINDING
}

MODULES = {
    TAXII_11_NS: tm11,
    TAXII_10_NS: tm10
}

SUPPORTED_CONTENT_BINDINGS = [
    XML_11_BINDING, XML_10_BINDING, CERT_EU_JSON_10_BINDING]

BINDINGS_TO_CONTENT_TYPE = {
    XML_11_BINDING: 'application/xml',
    XML_10_BINDING: 'application/xml',
    CERT_EU_JSON_10_BINDING: 'application/json'
}

TAXII_SERVICES_10 = 'urn:taxii.mitre.org:services:1.0'
TAXII_SERVICES_11 = 'urn:taxii.mitre.org:services:1.1'

BINDINGS_TO_SERVICES = {
    XML_10_BINDING: TAXII_SERVICES_10,
    XML_11_BINDING: TAXII_SERVICES_11,
    CERT_EU_JSON_10_BINDING: 'urn:taxii.mitre.org:services:1.0'
}

PROTOCOL_HTTP_10_BINDING = 'urn:taxii.mitre.org:protocol:http:1.0'
PROTOCOL_HTTPS_10_BINDING = 'urn:taxii.mitre.org:protocol:https:1.0'

SCHEMA_TO_PROTOCOL_BINDINGS = {
    'https': PROTOCOL_HTTPS_10_BINDING,
    'http': PROTOCOL_HTTP_10_BINDING
}

# Constant identifying a Status Message
MSG_STATUS_MESSAGE = 'Status_Message'
# Constant identifying a Discovery Request Message
MSG_DISCOVERY_REQUEST = 'Discovery_Request'
# Constant identifying a Discovery Response Message
MSG_DISCOVERY_RESPONSE = 'Discovery_Response'
# Constant identifying a Feed Information Request Message
MSG_FEED_INFORMATION_REQUEST = 'Feed_Information_Request'
# Constant identifying a Feed Information Response Message
MSG_FEED_INFORMATION_RESPONSE = 'Feed_Information_Response'
# Constant identifying a Subscription Management Request Message
MSG_MANAGE_FEED_SUBSCRIPTION_REQUEST = 'Subscription_Management_Request'
# Constant identifying a Subscription Management Response Message
MSG_MANAGE_FEED_SUBSCRIPTION_RESPONSE = 'Subscription_Management_Response'
# Constant identifying a Poll Request Message
MSG_POLL_REQUEST = 'Poll_Request'
# Constant identifying a Poll Response Message
MSG_POLL_RESPONSE = 'Poll_Response'
# Constant identifying a Inbox Message
MSG_INBOX_MESSAGE = 'Inbox_Message'

# New Message Types in TAXII 1.1

# Constant identifying a Status Message
MSG_POLL_FULFILLMENT_REQUEST = 'Poll_Fulfillment'
# Constant identifying a Collection Information Request
MSG_COLLECTION_INFORMATION_REQUEST = 'Collection_Information_Request'
# Constant identifying a Collection Information Response
MSG_COLLECTION_INFORMATION_RESPONSE = 'Collection_Information_Response'
# Constant identifying a Subscription Request
MSG_MANAGE_COLLECTION_SUBSCRIPTION_REQUEST = 'Subscription_Management_Request'
# Constant identifying a Subscription Response
MSG_MANAGE_COLLECTION_SUBSCRIPTION_RESPONSE = \
    'Subscription_Management_Response'


# Constant identifying a Service Type of Inbox
SVC_INBOX = 'INBOX'
# Constant identifying a Service Type of Poll
SVC_POLL = 'POLL'
# Constant identifying a Service Type of Feed Management
SVC_FEED_MANAGEMENT = 'FEED_MANAGEMENT'
# Constant identifying a Service Type of Discovery
SVC_DISCOVERY = 'DISCOVERY'

# Tuple of all TAXII 1.0 Service Types
SVC_TYPES_10 = (SVC_INBOX, SVC_POLL, SVC_FEED_MANAGEMENT, SVC_DISCOVERY)

# Renamed Status Types in TAXII 1.1
# Constant identifying a Service Type of Collection Management.
# "Feed Management" was renamed to "Collection Management" in TAXII 1.1.
SVC_COLLECTION_MANAGEMENT = 'COLLECTION_MANAGEMENT'


SVC_TYPES = [
    SVC_INBOX, SVC_POLL, SVC_COLLECTION_MANAGEMENT, SVC_DISCOVERY,
    SVC_FEED_MANAGEMENT]

# TAXII 1.0 Action Types

# Constant identifying an Action of Subscribe
ACT_SUBSCRIBE = 'SUBSCRIBE'
# Constant identifying an Action of Unsubscribe
ACT_UNSUBSCRIBE = 'UNSUBSCRIBE'
# Constant identifying an Action of Status
ACT_STATUS = 'STATUS'

# Tuple of all TAXII 1.0 Action Types
ACT_TYPES_10 = (ACT_SUBSCRIBE, ACT_UNSUBSCRIBE, ACT_STATUS)

# Constant identifying an Action of Pause
ACT_PAUSE = 'PAUSE'
# Constant identifying an Action of Resume
ACT_RESUME = 'RESUME'

# Subscription Status of Active
SS_ACTIVE = 'ACTIVE'
# Subscription Status of Paused
SS_PAUSED = 'PAUSED'
# Subscription Status of Unsubscribed
SS_UNSUBSCRIBED = 'UNSUBSCRIBED'


# Constant identifying a response type of Full
RT_FULL = 'FULL'
# Constant identifying a response type of Count only
RT_COUNT_ONLY = 'COUNT_ONLY'

# Constant identifying a collection type of Data Feed
CT_DATA_FEED = 'DATA_FEED'
# Constant identifying a collection type of Data Set
CT_DATA_SET = 'DATA_SET'

# Content Binding ID for STIX XML 1.0
CB_STIX_XML_10 = 'urn:stix.mitre.org:xml:1.0'
# Content Binding ID for STIX XML 1.0.1
CB_STIX_XML_101 = 'urn:stix.mitre.org:xml:1.0.1'
# Content Binding ID for STIX XML 1.1
CB_STIX_XML_11 = 'urn:stix.mitre.org:xml:1.1'
# Content Binding ID for STIX XML 1.1.1
CB_STIX_XML_111 = 'urn:stix.mitre.org:xml:1.1.1'
# Content Binding ID for CAP 1.1
CB_CAP_11 = 'urn:oasis:names:tc:emergency:cap:1.1'
# Content Binding ID for XML Encryption
CB_XENC_122002 = 'http://www.w3.org/2001/04/xmlenc#'
# Content Binding ID for SMIME
CB_SMIME = 'application/x-pks7-mime'



import logging

from . import constants as const


log = logging.getLogger(__name__)

SERVICE_TYPES = set(const.SVC_TYPES)


class Entity(object):
    '''Generic entity.'''
    raw = None


class ContentBlockCount(Entity):
    '''Content Block count.

    Represents an amount of content blocks in a poll query result.

    :param int count: amount of content blocks in a poll query result.
    :param bool is_partial:
           indicates whether the provided Record Count is the exact number
           of applicable records, or if the provided number is a lower bound
           and there may be more records than stated.
    '''
    def __init__(self, count, is_partial=False):
        self.count = count
        self.is_partial = is_partial

    def __repr__(self):
        t = '{cls}(count={count}, is_partial={is_partial})'
        return t.format(cls=type(self).__name__, **vars(self))


class Collection(Entity):
    '''Collection entity.

    Represents TAXII Collection and TAXII Feed objects.

    :param str name: name of the collection
    :param str description: description message
    :param str type: type of a collection. Supported values are
           :attr:`TYPE_FEED` and :attr:`TYPE_SET`
    :param bool available: if a collection marked as available
    :param list push_methods: availavle push methods for a collection,
           a list of :py:class:`cabby.entities.PushMethods`
    :param list content_bindings: a list of
           :py:class:`cabby.entities.ContentBinding`
    :param list polling_services: a list of
           :py:class:`cabby.entities.ServiceInstance`
    :param list subscription_methods: a list of
           :py:class:`cabby.entities.ServiceInstance`
    :param list receiving_inboxes: a list of
           :py:class:`cabby.entities.InboxService`
    :param int volume: collection's volume
    '''

    TYPE_FEED = const.CT_DATA_FEED
    TYPE_SET = const.CT_DATA_SET

    def __init__(self, name, description, type=TYPE_FEED, available=None,
                 push_methods=None, content_bindings=None,
                 polling_services=None, subscription_methods=None,
                 receiving_inboxes=None, volume=None):

        if type not in [self.TYPE_FEED, self.TYPE_SET]:
            log.error("Unknown collection type: %s", type)

        self.name = name
        self.description = description
        self.type = type

        self.available = available

        self.content_bindings = content_bindings or []
        self.push_methods = push_methods or []

        self.polling_services = polling_services or []
        self.subscription_methods = subscription_methods or []

        self.receiving_inboxes = receiving_inboxes or []
        self.volume = volume

    def __repr__(self):
        t = '{cls}(name={name}, type={type}, available={available})'
        return t.format(cls=type(self).__name__, **vars(self))


class ContentBinding(Entity):
    '''Content Binding entity.

    Represents TAXII Content Binding.

    :param str id: Content Binding ID
    :param list subtypes: Content Subtypes IDs
    '''

    def __init__(self, id, subtypes=None):
        self.id = id
        self.subtypes = subtypes or []

    def __repr__(self):
        t = '{cls}(id={id}, subtypes={subtypes})'
        return t.format(cls=type(self).__name__, **vars(self))


class ServiceInstance(Entity):
    '''Generic TAXII Service entity.

    :param str protocol: service Protocol Binding value
    :param str address: service network address
    :param list message_bindings: service Message Bindings,
                                  as list of strings
    '''

    def __init__(self, protocol, address, message_bindings):
        self.protocol = protocol
        self.address = address
        self.message_bindings = message_bindings

    def __repr__(self):
        t = '{cls}(protocol={protocol}, address={address})'
        return t.format(cls=type(self).__name__, **vars(self))


class InboxService(ServiceInstance):
    '''Inbox Service entity.

    Represents TAXII Inbox Service.

    :param str protocol: service Protocol Binding value
    :param str address: service network address
    :param list message_bindings: service Message Bindings, as list of strings
    :param list content_bindings: a list of
                :py:class:`cabby.entities.ContentBinding`
    '''

    def __init__(self, protocol, address, message_bindings,
                 content_bindings=None):

        super(InboxService, self).__init__(protocol, address, message_bindings)

        self.content_bindings = content_bindings or []


class PushMethod(Entity):
    '''Push Method entity.

    Represents TAXII Push Method.

    :param str protocol: service Protocol Binding value
    :param list message_bindings: service Message Bindings, as list of strings
    '''

    def __init__(self, protocol, message_bindings):
        self.protocol = protocol
        self.message_bindings = message_bindings

    def __repr__(self):
        t = '{cls}(protocol={protocol})'
        return t.format(cls=type(self).__name__, **vars(self))


class SubscriptionParameters(Entity):
    '''Subscription Parameters Entity.

    Represents TAXII Subscription Parameters.

    :param str response_type: response type. Supported values are
               :attr:`TYPE_FULL` and :attr:`TYPE_COUNTS`
    :param list content_bindings: a list of
               :py:class:`cabby.entities.ContentBinding`
    '''

    TYPE_FULL = const.RT_FULL
    TYPE_COUNT = const.RT_COUNT_ONLY

    def __init__(self, response_type, content_bindings=None):
        if response_type not in [self.TYPE_FULL, self.TYPE_COUNT]:
            log.error("Unknown response type: %s", response_type)
        self.response_type = response_type
        self.content_bindings = content_bindings or []

    def __repr__(self):
        t = '{cls}(response_type={response_type})'
        return t.format(cls=type(self).__name__, **vars(self))


class DetailedServiceInstance(Entity):
    '''Detailed description of a generic TAXII Service instance

    :param str type: service type. Supported values are
                        in :py:data:`cabby.entities.SERVICE_TYPES`
    :param str version: service version. Supported values are
                        :attr:`VERSION_10` and :attr:`VERSION_11`
    :param str protocol: service Protocol Binding value
    :param str address: service network address
    :param list message_bindings: service Message Bindings, as list of strings
    :param bool available: if service is marked as available
    :param str message: message attached to a service
    '''

    VERSION_10 = const.TAXII_SERVICES_10
    VERSION_11 = const.TAXII_SERVICES_11

    PROTOCOL_HTTP = const.PROTOCOL_HTTP_10_BINDING
    PROTOCOL_HTTPS = const.PROTOCOL_HTTPS_10_BINDING

    def __init__(self, type, version, protocol, address, message_bindings,
                 available=None, message=None):

        if version not in [self.VERSION_10, self.VERSION_11]:
            log.error("Unknown service version: %s", version)
        if protocol not in [self.PROTOCOL_HTTP, self.PROTOCOL_HTTPS]:
            log.error("Unknown service protocol: %s", protocol)

        self.type = type
        self.version = version
        self.protocol = protocol
        self.address = address
        self.message_bindings = message_bindings

        self.available = available
        self.message = message

    def __repr__(self):
        t = '{cls}(type={type}, address={address})'
        return t.format(cls=type(self).__name__, **vars(self))


class InboxDetailedService(DetailedServiceInstance):
    '''Detailed description of TAXII Inbox Service.

    :param str type: service type. Supported values are
                     in :py:data:`cabby.entities.SERVICE_TYPES`
    :param str version: service version. Supported values are
                        :attr:`VERSION_10` and :attr:`VERSION_11`
    :param str protocol: service Protocol Binding value
    :param str address: service network address
    :param list message_bindings: service Message Bindings, as list of strings
    :param list content_bindings: a list of
                                  :py:class:`cabby.entities.ContentBinding`
    :param bool available: if service is marked as available
    :param str message: message attached to a service
    '''

    def __init__(self, content_bindings, **kwargs):
        super(InboxDetailedService, self).__init__(**kwargs)
        self.content_bindings = content_bindings


class ContentBlock(Entity):
    '''Content Block entity.

    Represents TAXII Content Block.

    :param str content: TAXII message payload
    :param `cabby.entities.ContentBinding` content_binding: Content Binding
    :param datetime timestamp: content block timestamp label
    '''

    def __init__(self, content, content_binding, timestamp):
        self.content = content
        self.binding = content_binding
        self.timestamp = timestamp

    def __repr__(self):
        t = '{cls}(timestamp={timestamp})'
        return t.format(cls=type(self).__name__, **vars(self))


class SubscriptionResponse(Entity):
    '''Subscription Response entity.

    :param str collection_name: collection name
    :param str message: message attached to Subscription Response
    :param list subscriptions: a list of `cabby.entities.Subscription`
    '''

    def __init__(self, collection_name, message=None, subscriptions=None):
        self.collection_name = collection_name
        self.message = message
        self.subscriptions = subscriptions or []

    def __repr__(self):
        t = '{cls}(collection_name={collection_name})'
        return t.format(cls=type(self).__name__, **vars(self))


class Subscription(Entity):
    '''Subscription entity.

    :param str subscription_id: subscription ID
    :param str status: subscription status. Supported values are
                 :attr:`STATUS_UNKNOWN`, :attr:`STATUS_ACTIVE`,
                 :attr:`STATUS_PAUSED`, :attr:`STATUS_UNSUBSCRIBED`
    :param list delivery_parameters: a list of `cabby.entities.InboxService`
    :param list subscription_parameters:
                a list of `cabby.entities.SubscriptionParameters`
    :param list poll_instances: a list of `cabby.entities.ServiceInstance`
    '''
    STATUS_UNKNOWN = 'UNKNOWN'
    STATUS_ACTIVE = const.SS_ACTIVE
    STATUS_PAUSED = const.SS_PAUSED
    STATUS_UNSUBSCRIBED = const.SS_UNSUBSCRIBED

    def __init__(self, subscription_id, status=STATUS_UNKNOWN,
                 delivery_parameters=None, subscription_parameters=None,
                 poll_instances=None):

        assert status in [self.STATUS_UNKNOWN, self.STATUS_ACTIVE,
                          self.STATUS_PAUSED, self.STATUS_UNSUBSCRIBED]

        self.id = subscription_id
        self.status = status
        self.delivery_parameters = delivery_parameters
        self.subscription_parameters = subscription_parameters
        self.poll_instances = poll_instances

    def __repr__(self):
        t = '{cls}(subscription_id={id}, status={status})'
        return t.format(cls=type(self).__name__, **vars(self))





from cabby.constants import (
    SVC_COLLECTION_MANAGEMENT, SVC_FEED_MANAGEMENT)

from .commons import run_client, get_basic_arg_parser


def _runner(client, path, args):

    collections = []

    if path:
        collections.extend(client.get_collections(uri=path))
    else:
        service_types = [SVC_COLLECTION_MANAGEMENT, SVC_FEED_MANAGEMENT]
        for service in client.get_services(service_types=service_types):
            collections.extend(
                client.get_collections(uri=service.address))

    for c in collections:
        if args.as_xml:
            print(c.raw.to_xml(pretty_print=True))
        else:
            print(c.raw.to_text())


def fetch_collections():
    run_client(get_basic_arg_parser(), _runner)



from furl import furl
import logging

import libtaxii

from . import dispatcher, utils
from .converters import to_detailed_service_instance_entity
from .exceptions import (
    AmbiguousServicesError,
    ClientException,
    NoURIProvidedError,
    ServiceNotFoundError,
    UnsuccessfulStatusError,
)
from six.moves import map


class AbstractClient(object):
    '''
    Abstract client class.

    This class can not be used directly, use :py:meth:`cabby.create_client`
    to create client instances.
    '''

    SUPPORTED_SCHEMES = ['http', 'https']

    taxii_version = None

    def __init__(self, host=None, discovery_path=None, port=None,
                 use_https=False, headers=None, timeout=None):

        self.host = host
        self.port = port
        self.use_https = use_https

        self.discovery_path = discovery_path
        self.services = None

        self.proxies = None
        self.verify_ssl = True
        self.ca_cert = None
        self.cert_file = None
        self.key_file = None
        self.key_password = None

        self.username = None
        self.password = None
        self.jwt_url = None

        self.jwt_token = None
        self.headers = headers or {}
        self.timeout = timeout

        self.log = logging.getLogger(
            "{}.{}".format(self.__module__, self.__class__.__name__))

    def set_auth(self, ca_cert=None, cert_file=None, key_file=None,
                 key_password=None, username=None, password=None,
                 jwt_auth_url=None, verify_ssl=True):
        '''
        Set authentication credentials.

        ``jwt_auth_url`` is required for JWT based authentication. If
        it is not specified but ``username`` and ``password`` are provided,
        client will configure Basic authentication.

        SSL authentication can be combined with JWT and Basic
        authentication.

        :param str ca_cert: a path to CA SSL certificate file
        :param str cert_file: a path to SSL certificate file
        :param str key_file: a path to SSL key file
        :param str username: username, used in basic auth or JWT auth
        :param str password: password, used in basic auth or JWT auth
        :param str key_password: same argument as in
            ``ssl.SSLContext.load_cert_chain`` - may be a function to call
            to get the password for decrypting the private key or
            string/bytes/bytearray. It will only be called if the private
            key is encrypted and a password is necessary.
        :param str jwt_auth_url: URL used to obtain JWT token
        :param bool/str verify_ssl: set to False to skip checking host's SSL
            certificate. Set to True to check certificate against public CAs or
            set to filepath to check against custom CA bundle.
        '''

        self.ca_cert = ca_cert
        self.cert_file = cert_file
        self.key_file = key_file
        self.key_password = key_password

        self.username = username
        self.password = password

        if jwt_auth_url:
            self.jwt_url = self._prepare_url(jwt_auth_url)

        self.verify_ssl = verify_ssl

    def set_proxies(self, proxies):
        '''
        Set proxy properties.

        Cause requests to go through a proxy.
        Must be a dictionary mapping protocol names to URLs of proxies.

        :param dir proxies: dictionary mapping protocol names to URLs
        '''
        self.proxies = proxies

    def _prepare_url(self, uri):

        fu = furl(uri)

        if fu.scheme and fu.scheme not in self.SUPPORTED_SCHEMES:
            raise ValueError(
                'Scheme "{}" is not supported. Use one of: {}'
                .format(fu.scheme, ', '.join(self.SUPPORTED_SCHEMES)))

        use_https = self.use_https or (fu.scheme == 'https')

        fu.scheme = fu.scheme or ('https' if use_https else 'http')
        fu.host = fu.host or self.host
        fu.port = self.port if self.port else (
            fu.port or (443 if use_https else 80))

        if not fu.host:
            raise ValueError('Host name is not provided: {}'.format(fu.url))

        return fu.url

    def refresh_jwt_token(self, session=None):
        '''
        Obtain JWT token using provided JWT session,
        url, username and password.
        '''
        session = session or self.prepare_generic_session()
        self.jwt_token = dispatcher.obtain_jwt_token(
            session,
            self._prepare_url(self.jwt_url),
            self.username,
            self.password)
        session.auth = dispatcher.JWTAuth(self.jwt_token)
        return self.jwt_token

    def prepare_generic_session(self):
        '''
        Prepare basic generic session with configured
        proxies, headers, username/password (if no JWT url configured),
        cert file, key file and SSL verification flags.
        '''
        return dispatcher.get_generic_session(
            proxies=self.proxies,
            headers=self.headers,
            username=self.username if not self.jwt_url else None,
            password=self.password if not self.jwt_url else None,
            cert_file=self.cert_file,
            key_file=self.key_file,
            key_password=self.key_password,
            ca_cert=self.ca_cert,
            verify_ssl=self.verify_ssl,
            jwt_token=self.jwt_token,
        )

    def _execute_request(self, request, uri=None, service_type=None):
        '''
        Execute generic TAXII request.

        A service is defined by ``uri`` parameter or is chosen from pre-cached
        services by ``service_type``.
        '''
        if not uri and not service_type:
            raise NoURIProvidedError('URI or service_type needed')
        elif not uri:
            service = self._get_service(service_type)
            uri = service.address

        if (self.key_file
                and not self.key_password
                and utils.if_key_encrypted(self.key_file)):
            raise ValueError(
                'Key file is encrypted but key password was not provided')

        session = self.prepare_generic_session()

        if self.jwt_url and self.username and self.password:
            if not self.jwt_token:
                self.refresh_jwt_token(session=session)

        for attempt in (1, 2):
            try:
                return dispatcher.send_taxii_request(
                    session,
                    self._prepare_url(uri),
                    request,
                    taxii_binding=self.taxii_binding,
                    timeout=self.timeout)
            except UnsuccessfulStatusError as exc:
                # Refresh the token once if authorization failed and retry
                if attempt == 1 and exc.status == libtaxii.ST_UNAUTHORIZED:
                    self.refresh_jwt_token(session=session)
                    continue
                raise

    def _generate_id(self):
        return libtaxii.common.generate_message_id(version=self.services_version)

    def _get_service(self, service_type):
        candidates = self.get_services(service_type=service_type)

        if not candidates:
            raise ServiceNotFoundError(
                "Service with type '{}' is not advertised"
                .format(service_type))

        elif len(candidates) > 1:
            raise AmbiguousServicesError(
                "{} services found with type '{}'. Specify the exact URI"
                .format(len(candidates), service_type))

        return candidates[0]

    def get_services(self, service_type=None, service_types=None):
        '''
        Get services advertised by TAXII server.

        This method will try to do automatic discovery by calling
        :py:meth:`discover_services`.

        :param str service_type: filter services by specific type. Accepted
                                 values are listed in
                                 :py:data:`cabby.entities.SERVICE_TYPES`
        :param str service_types: filter services by multiple types. Accepted
                                 values are listed in
                                 :py:data:`cabby.entities.SERVICE_TYPES`

        :return: list of service instances
        :rtype: list of :py:class:`cabby.entities.DetailedServiceInstance`
                (or :py:class:`cabby.entities.InboxDetailedService`)

        :raises ValueError:
                if URI provided is invalid or schema is not supported
        :raises `cabby.exceptions.HTTPError`:
                if HTTP error happened
        :raises `cabby.exceptions.UnsuccessfulStatusError`:
                if Status Message received and status_type is not `SUCCESS`
        :raises `cabby.exceptions.ServiceNotFoundError`:
                if no service found
        :raises `cabby.exceptions.AmbiguousServicesError`:
                more than one service with type specified
        :raises `cabby.exceptions.NoURIProvidedError`:
                no URI provided and client can't discover services
        '''
        if self.services:
            services = self.services
        else:
            try:
                services = self.discover_services()
            except ClientException as e:
                self.log.error('Can not autodiscover advertised services')
                raise e

        if service_type:
            return [s for s in services if s.type == service_type]
        elif service_types:
            return [s for s in services if s.type in service_types]
        else:
            return services

    def discover_services(self, uri=None, cache=True):
        '''
        Discover services advertised by TAXII server.

        This method will send discovery request to a service, defined
        by ``uri`` or constructor's connection parameters.

        :param str uri: URI path to a specific TAXII service
        :param bool cache: if discovered services should be cached

        :return: list of TAXII services
        :rtype: list of :py:class:`cabby.entities.DetailedServiceInstance`
                (or :py:class:`cabby.entities.InboxDetailedService`)

        :raises ValueError:
                if URI provided is invalid or schema is not supported
        :raises `cabby.exceptions.HTTPError`:
                if HTTP error happened
        :raises `cabby.exceptions.UnsuccessfulStatusError`:
                if Status Message received and status_type is not `SUCCESS`
        :raises `cabby.exceptions.ServiceNotFoundError`:
                if no Discovery servicefound
        :raises `cabby.exceptions.AmbiguousServicesError`:
                more than one service with type specified
        :raises `cabby.exceptions.NoURIProvidedError`:
                no URI provided and client can't discover services
        '''

        uri = uri or self.discovery_path

        if not uri:
            raise NoURIProvidedError('Discovery service URI is not specified')

        response = self._discovery_request(uri)

        services = list(map(
            to_detailed_service_instance_entity,
            response.service_instances))

        self.log.info("%d services discovered", len(services))

        if cache:
            self.services = services

        return services

    def __repr__(self):
        t = '{name}(host={host}, port={port}, discovery_path={discovery_path})'
        return t.format(
            name=type(self).__name__,
            host=self.host,
            port=self.port,
            discovery_path=self.discovery_path,
        )




import logging
import sys
import argparse
from colorlog import ColoredFormatter

from .. import create_client

logging.captureWarnings(True)

log = logging.getLogger(__name__)

TAXII_10 = '1.0'
TAXII_11 = '1.1'
DEFAULT_VERSION = TAXII_11
VERSION_CHOICES = [TAXII_10, TAXII_11]


def get_basic_arg_parser():

    parser = argparse.ArgumentParser(
        description="TAXII client",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "--host", dest="host",
        help="host where the TAXII Service is hosted")

    parser.add_argument(
        "--port", dest="port",
        type=int, help="port where the TAXII Service is hosted")

    parser.add_argument(
        "--discovery", dest="discovery",
        help="path to a TAXII Discovery service")

    parser.add_argument(
        "--path", dest="uri",
        help=("absolute path (as URL) or relative path "
              "to a specific TAXII Service"))

    parser.add_argument(
        "--https", dest="https",
        action='store_true',
        help="if the client should use HTTPS")

    parser.add_argument(
        "--verify", dest="verify",
        help=('Set to "no" to skip checking the host\'s SSL certificate.'
              ' You can also pass the path to a CA file for private certs.'
              ' Defaults to "yes".'))

    parser.add_argument(
        "--timeout", dest="timeout",
        type=int,
        help="HTTP request timeout in seconds")

    parser.add_argument(
        "--ca-cert", dest="ca_cert", help="CA certificate file")
    parser.add_argument("--cert", dest="cert", help="certificate file")
    parser.add_argument("--key", dest="key", help="private key file")
    parser.add_argument(
        "--key-password", dest="key_password",
        help="private key password")

    parser.add_argument(
        "--username", dest="username",
        help="username for authentication")

    parser.add_argument(
        "--password", dest="password",
        help="password for authentication")

    parser.add_argument(
        "--jwt-auth", dest="jwt_auth_url",
        help="JWT authentication URL")

    parser.add_argument(
        "--proxy-url", dest="proxy_url",
        help=("proxy address formatted as URL. Can be set to 'noproxy' "
              "to force library to not use any proxy"))

    parser.add_argument(
        "--proxy-type", dest="proxy_type",
        choices=['http', 'https'],
        help="proxy type")

    parser.add_argument(
        "--header", dest="headers",
        action='append',
        help="header to send with the request, as header:value pair")

    parser.add_argument(
        "-v", "--verbose", dest="verbose",
        action='store_true',
        help="verbose mode")

    parser.add_argument(
        "-x", "--as-xml", dest="as_xml",
        action='store_true',
        help="output raw XML")

    parser.add_argument(
        "-t", "--taxii-version", dest="version",
        default=DEFAULT_VERSION,
        choices=VERSION_CHOICES,
        help="TAXII version to use")

    return parser


def is_args_valid(args):

    if not args.uri and not args.discovery:
        log.error("Either Discovery Service path '--discovery'"
                  " or specific service path '--path' need to be provided")
        return False

    return True


def prepare_headers(raw_headers):
    headers = {}

    for line in raw_headers:
        header, value = line.split(":", 1)
        headers[header] = value.strip('"\'')

    return headers


def run_client(parser, run_func):

    args = parser.parse_args()

    level = logging.DEBUG if args.verbose else logging.INFO
    configure_color_logging(level=level)

    if not is_args_valid(args):
        sys.exit(1)

    headers = prepare_headers(args.headers) if args.headers else None

    client = create_client(host=args.host, discovery_path=args.discovery,
                           port=args.port, use_https=args.https,
                           version=args.version, headers=headers)

    if args.verify == "yes":
        verify_ssl = True
    elif args.verify == "no":
        verify_ssl = False
    elif args.verify:
        verify_ssl = args.verify
    else:
        verify_ssl = True

    client.set_auth(
        ca_cert=args.ca_cert,
        cert_file=args.cert,
        key_file=args.key,
        username=args.username,
        password=args.password,
        key_password=args.key_password,
        jwt_auth_url=args.jwt_auth_url,
        verify_ssl=verify_ssl
    )

    client.timeout = args.timeout

    if args.proxy_url:
        client.set_proxies({args.proxy_type: args.proxy_url})

    try:
        run_func(client, args.uri, args)
    except Exception as e:
        log.error(e, exc_info=args.verbose)


def configure_color_logging(level, logger_name=None):

    logger = logging.getLogger(logger_name)
    logger.setLevel(level)

    if level == logging.DEBUG:
        format_string = ("%(asctime)s %(name)s "
                         "%(log_color)s%(levelname)s:%(reset)s "
                         "%(white)s%(message)s")
    else:
        format_string = ("%(asctime)s "
                         "%(log_color)s%(levelname)s:%(reset)s "
                         "%(white)s%(message)s")

    formatter = ColoredFormatter(
        format_string,
        datefmt=None,
        reset=True,
        log_colors={
            'DEBUG': 'cyan',
            'INFO': 'green',
            'WARNING': 'yellow',
            'ERROR': 'red',
            'CRITICAL': 'red',
        }
    )

    handlers = logging.StreamHandler(sys.stderr)
    handlers.setFormatter(formatter)
    logger.addHandler(handlers)



'''
    Cabby, python library for interacting with TAXII servers.
'''
from ._version import __version__  # noqa: used in setup.py

from .client10 import Client10
from .client11 import Client11


try:
    from urllib.parse import urlparse
except ImportError:     # Python 2.7
    from urlparse import urlparse


def create_client(host=None, port=None, discovery_path=None, use_https=False,
                  discovery_url=None, version="1.1", headers=None):
    '''Create a client instance (TAXII version specific).

    ``host``, ``port``, ``use_https``, ``discovery_path`` values
    can be overridden per request with ``uri`` argument passed to
    a client's method.

    :param str host: TAXII server hostname
    :param int port: TAXII server port
    :param str discovery_path: Discovery Service relative path
    :param bool use_https: if HTTPS should be used
    :param string discovery_url: URL to infer host, port, discovery_path,
                                 and use_https.
    :param string version: TAXII version (1.1 or 1.0)
    :param dict headers: additional headers to pass with TAXII messages

    :return: client instance
    :rtype: :py:class:`cabby.client11.Client11` or
            :py:class:`cabby.client10.Client10`
    '''
    if discovery_url:
        parsed = urlparse(discovery_url)
        if not host and parsed.hostname:
            host = parsed.hostname
        if not port and parsed.port:
            port = parsed.port
        if not discovery_path and parsed.path:
            discovery_path = parsed.path
        if parsed.scheme:
            use_https = parsed.scheme == 'https'

    params = dict(
        host=host,
        port=port,
        use_https=use_https,
        discovery_path=discovery_path,
        headers=headers)

    if version == '1.1':
        return Client11(**params)
    elif version == '1.0':
        return Client10(**params)
    else:
        raise ValueError("TAXII version %s is not supported" % version)



from collections import namedtuple
import json
import os
import ssl
import sys
import logging

from six import StringIO

import furl
import gzip
import requests
from lxml import etree
from six.moves import urllib
from requests.auth import AuthBase, HTTPBasicAuth

from libtaxii import messages_11 as tm11
from libtaxii import messages_10 as tm10

from . import constants as const
from ._version import __version__ as cabby_version

from .exceptions import (
    UnsuccessfulStatusError, HTTPError, InvalidResponseError)


log = logging.getLogger(__name__)


def raise_http_error(status_code, response_stream=None):
    if log.isEnabledFor(logging.DEBUG) and response_stream:
        body = response_stream.read()
        log.debug("Response:\n%s", body.decode('utf-8'))
    raise HTTPError(status_code)


def send_taxii_request(
        session, url, request, taxii_binding=None, timeout=None):
    '''
    Send XML message to a TAXII service and parse a response.
    '''

    log.info("Sending {} to {}".format(request.message_type, url))

    request_body = request.to_xml(pretty_print=True)

    log.debug("Request:\n%s", request_body.decode('utf-8'))

    session = get_taxii_session(
        session,
        url_scheme=furl.furl(url).scheme,
        message_binding=taxii_binding)

    stream, headers = request_stream(session, url, request_body, timeout)

    gen = _parse_response(stream, headers, version=request.version)
    obj = next(gen)

    if obj == const.STREAM_MARKER:
        return gen
    elif hasattr(obj, 'status_type'):
        if obj.status_type != 'SUCCESS':
            raise UnsuccessfulStatusError(obj)
        else:
            return None
    return obj


def request_stream(session, url, request_body, timeout, headers=None):
    if session._cabby_key_password:
        # Workaround until
        # https://github.com/kennethreitz/requests/issues/2519 is fixed
        try:
            response = request_with_key_password(
                session, url, request_body, timeout, headers)
        except urllib.error.HTTPError as e:
            log.error(
                "Error while connecting to {}".format(url),
                exc_info=True)
            raise_http_error(e.getcode())

        stream, headers = response, response.headers
    else:
        response = session.post(
            url,
            data=request_body,
            stream=True,
            timeout=timeout,
            headers=headers)
        if not response.ok:
            raise_http_error(response.status_code, response.raw)

        stream, headers = response.raw, response.headers

    if 'gzip' in headers.get('content-encoding', ''):

        if sys.version_info < (3, 2):
            # It is impossible to seek on a HTTP response inside gzipped stream
            # in gzip lib for python < v3.2, so we are forced to read all
            # stream content into memory
            stream = StringIO(stream.read())

        stream = gzip.GzipFile(fileobj=stream)

    return stream, headers


def _cleanup_batch(curr_elem, batch):
    current_parent = curr_elem.getparent()

    for elem in batch:
        if elem.getparent() is not None:
            parent = elem.getparent()
        else:
            parent = current_parent
        parent.remove(elem)

    del batch[:]


def _stream_poll_response(namespace, stream):

    module = const.MODULES[namespace]

    response_cls = const.MODULES[namespace].PollResponse

    batch_max_size = 3
    to_delete_batch = []

    for action, elem in stream:
        if action == 'end':
            tag = elem.xpath('local-name()')

            # If current element is ContentBlock
            if tag == module.ContentBlock.NAME:
                obj = module.ContentBlock.from_etree(elem)

            # If current element is PollResponse
            # meaning that this is a last one
            elif tag == response_cls.message_type:
                _cleanup_batch(elem, to_delete_batch)
                obj = response_cls.from_etree(elem)
            else:
                continue

            if log.isEnabledFor(logging.DEBUG):
                log.debug("Stream element:\n{}"
                          .format(etree.tostring(elem).decode('utf-8')))

            yield obj

            # Cleaning up element to free up memory
            elem.clear()

            # Removing all elements from a batch if it is time
            if len(to_delete_batch) >= batch_max_size:
                _cleanup_batch(elem, to_delete_batch)

            # Postponing removal of the element from a tree
            # to avoid memory corruption (libxml2 crashes)
            to_delete_batch.append(elem)


def _parse_response(stream, headers, version):

    content_type = headers.get('X-TAXII-Content-Type')

    if not content_type:
        body = stream.read()
        headers = '\n'.join([
            '{}={}'.format(k, v) for k, v in headers.items()])
        log.debug(
            "Invalid response:\n{}\n{}".format(headers, body))
        raise InvalidResponseError("Invalid response received")

    elif content_type not in const.SUPPORTED_CONTENT_BINDINGS:
        raise ValueError('Unsupported X-TAXII-Content-Type: {}'
                         .format(content_type))
    elif content_type == const.CERT_EU_JSON_10_BINDING:
        body = stream.read()
        yield tm10.get_message_from_json(body)

    support_huge_trees = not os.environ.get('CABBY_NO_HUGE_TREES')
    gen = etree.iterparse(
        stream,
        events=('start', 'end'),
        # inject default attributes from DTD or XMLSchema
        attribute_defaults=False,
        # validate against a DTD referenced by the document
        dtd_validation=False,
        # use DTD for parsing
        load_dtd=False,
        # prevent network access for related files (default: True)
        no_network=True,
        # try hard to parse through broken XML
        recover=False,
        # discard blank text nodes that appear ignorable
        remove_blank_text=False,
        # discard comments
        remove_comments=False,
        # discard processing instructions
        remove_pis=False,
        # replace CDATA sections by normal text content (default: True)
        strip_cdata=True,
        # save memory for short text content (default: True)
        compact=True,
        # replace entities by their text value (default: True)
        resolve_entities=False,
        # enable/disable security restrictions and support very deep
        # trees and very long text content
        huge_tree=support_huge_trees)

    action, root = next(gen)
    namespace = etree.QName(root).namespace
    message_type = root.xpath('local-name()')

    if namespace not in const.VERSIONS:
        raise ValueError(
            'Unsupported namespace: {}'.format(namespace))
    elif version != const.VERSIONS[namespace]:
        raise InvalidResponseError(
            "Response TAXII version '{}' "
            "does not match request version '{}'"
            .format(const.VERSIONS[namespace], version))

    if message_type in [tm11.PollResponse.message_type,
                        tm10.PollResponse.message_type]:

        yield const.STREAM_MARKER

        for obj in _stream_poll_response(namespace, gen):
            yield obj

    else:
        # Walk tree iterator to the end to fill up root element
        for _ in gen:
            pass

        if log.isEnabledFor(logging.DEBUG):
            log.debug("Response:\n%s",
                      etree.tostring(root, pretty_print=True).decode('utf-8'))

        yield _parse_full_tree(content_type, message_type, root)


def _parse_full_tree(content_type, message_type, elem):

    if content_type == const.XML_10_BINDING:
        tm = tm10
    elif content_type == const.XML_11_BINDING:
        tm = tm11
    else:
        raise ValueError(
            "Unsupported content type binding {}".format(content_type))

    if message_type == const.MSG_DISCOVERY_REQUEST:
        return tm.DiscoveryRequest.from_etree(elem)
    elif message_type == const.MSG_DISCOVERY_RESPONSE:
        return tm.DiscoveryResponse.from_etree(elem)
    elif message_type == const.MSG_POLL_REQUEST:
        return tm.PollRequest.from_etree(elem)
    elif message_type == const.MSG_POLL_RESPONSE:
        return tm.PollResponse.from_etree(elem)
    elif message_type == const.MSG_STATUS_MESSAGE:
        return tm.StatusMessage.from_etree(elem)
    elif message_type == const.MSG_INBOX_MESSAGE:
        return tm.InboxMessage.from_etree(elem)

    if tm == tm11:
        # TAXII 1.1 specific
        if message_type == const.MSG_COLLECTION_INFORMATION_REQUEST:
            return tm.CollectionInformationRequest.from_etree(elem)
        elif message_type == const.MSG_COLLECTION_INFORMATION_RESPONSE:
            return tm.CollectionInformationResponse.from_etree(elem)
        elif message_type == const.MSG_MANAGE_COLLECTION_SUBSCRIPTION_REQUEST:
            return tm.ManageCollectionSubscriptionRequest.from_etree(elem)
        elif message_type == const.MSG_MANAGE_COLLECTION_SUBSCRIPTION_RESPONSE:
            return tm.ManageCollectionSubscriptionResponse.from_etree(elem)
        elif message_type == const.MSG_POLL_FULFILLMENT_REQUEST:
            return tm.PollFulfillmentRequest.from_etree(elem)
    elif tm == tm10:
        # TAXII 1.0 specific
        if message_type == const.MSG_FEED_INFORMATION_REQUEST:
            return tm.FeedInformationRequest.from_etree(elem)
        elif message_type == const.MSG_FEED_INFORMATION_RESPONSE:
            return tm.FeedInformationResponse.from_etree(elem)
        elif message_type == const.MSG_MANAGE_FEED_SUBSCRIPTION_REQUEST:
            return tm.ManageFeedSubscriptionRequest.from_etree(elem)
        elif message_type == const.MSG_MANAGE_FEED_SUBSCRIPTION_RESPONSE:
            return tm.ManageFeedSubscriptionResponse.from_etree(elem)

    raise ValueError('Unknown message_type: %s' % message_type)


class JWTAuth(AuthBase):

    """Attaches JWT Authentication to the given Request object."""
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers['Authorization'] = 'Bearer {}'.format(self.token)
        return r


def get_generic_session(
    proxies=None,
    headers=None,
    username=None,
    password=None,
    cert_file=None,
    key_file=None,
    key_password=None,
    ca_cert=None,
    verify_ssl=True,
    jwt_token=None,
):
    session = requests.Session()
    if ca_cert:
        session.verify = ca_cert
    else:
        session.verify = verify_ssl
    if proxies:
        session.proxies = proxies
    if headers:
        session.headers.update(headers)
    session.headers['User-Agent'] = 'Cabby {}'.format(cabby_version)
    if username and password:
        session.auth = HTTPBasicAuth(username, password)
    if cert_file and key_file:
        session.cert = (cert_file, key_file)
    if jwt_token:
        session.auth = JWTAuth(jwt_token)
    session._cabby_key_password = key_password
    return session


def set_jwt_token(session, jwt_token):
    session.auth = JWTAuth(jwt_token)
    return session


def get_taxii_session(session, url_scheme='https', content_type=None,
                      message_binding=const.XML_11_BINDING,
                      service_binding=None):

    if not content_type:
        if message_binding not in const.BINDINGS_TO_CONTENT_TYPE:
            raise ValueError('No content type provided')
        content_type = const.BINDINGS_TO_CONTENT_TYPE[message_binding]

    if not service_binding:
        if message_binding not in const.BINDINGS_TO_SERVICES:
            raise ValueError('No service binding provided')
        service_bindings = const.BINDINGS_TO_SERVICES[message_binding]

    if url_scheme not in const.SCHEMA_TO_PROTOCOL_BINDINGS:
        raise ValueError(
            'No known protocol bindings for scheme {}'.format(url_scheme))

    session.headers.update({
        'Content-Type': content_type,
        'Accept': content_type,
        'X-TAXII-Content-Type': message_binding,
        'X-TAXII-Accept': message_binding,
        'X-TAXII-Services': service_bindings,
        'X-TAXII-Protocol': const.SCHEMA_TO_PROTOCOL_BINDINGS[url_scheme]
    })
    return session


def obtain_jwt_token(session, jwt_url, username, password, timeout=None):
    log.info("Obtaining JWT token from {}".format(jwt_url))

    request_data = json.dumps({'username': username, 'password': password})
    request_body = request_data.encode('utf-8')
    headers = {'Content-Type': 'application/json'}

    stream, headers = request_stream(
        session, jwt_url, request_body, timeout, headers)
    response_body = stream.read().decode('utf-8')
    if not response_body:
        raise ValueError("empty response")
    response_data = json.loads(response_body)

    if 'token' not in response_data:
        log.debug("Incorrect JWT response:\n{}".format(response_body))
        raise ValueError("No token found in JWT auth response")
    return response_data['token']


def request_with_key_password(
        session, url, request_body, timeout=None, headers=None):
    if sys.version_info < (2, 7, 9):
        raise ValueError(
            'Key password specification is not supported in Python < v2.7.9')

    if session.auth:
        # Using Requests Session's auth handlers to fill in proper headers
        DummyRequest = namedtuple('DummyRequest', ['headers'])
        request_headers = session.auth(
            DummyRequest(headers=session.headers)).headers
    else:
        request_headers = session.headers
    if headers:
        request_headers.update(headers)

    # Take the TLS details from the session object and use them with urllib.
    # See also 'get_generic_session' which sets many of these attributes.

    # session 'verify' attribute can be a bool or a path to a CA bundle:
    ca_cert = None
    if not isinstance(session.verify, bool):
        ca_cert = session.verify
    context = ssl.create_default_context(
        ssl.Purpose.CLIENT_AUTH, cafile=ca_cert)

    cert_file, key_file = session.cert
    key_password = session._cabby_key_password
    context.load_cert_chain(cert_file, key_file, password=key_password)

    if session.verify:
        context.verify_mode = ssl.CERT_REQUIRED
        if not ca_cert:
            context.set_default_verify_paths()
    else:
        context.verify_mode = ssl.CERT_NONE

    handlers = [urllib.request.HTTPSHandler(context=context)]
    if session.proxies:
        handlers.append(urllib.request.ProxyHandler(session.proxies))

    opener = urllib.request.build_opener(*handlers)
    request = urllib.request.Request(url, request_body, request_headers)

    if timeout:
        return opener.open(request, timeout=timeout)
    else:
        return opener.open(request)




import libtaxii.messages_11 as tm11

from . import constants as const
from .abstract import AbstractClient
from .entities import ContentBlockCount
from .converters import (
    to_subscription_response_entity, to_content_block_entity,
    to_collection_entities
)
from .utils import (
    pack_content_bindings, get_utc_now, pack_content_binding
)


class Client11(AbstractClient):
    '''Client implementation for TAXII Specification v1.1

    Use :py:meth:`cabby.create_client` to create client instances.
    '''

    taxii_binding = const.XML_11_BINDING
    services_version = const.TAXII_SERVICES_11

    def _discovery_request(self, uri):
        request = tm11.DiscoveryRequest(message_id=self._generate_id())
        response = self._execute_request(request, uri=uri)
        return response

    def __subscription_status_request(self, action, collection_name,
                                      subscription_id=None, uri=None):

        request_params = dict(
            message_id=self._generate_id(),
            action=action,
            collection_name=collection_name,
            subscription_id=subscription_id
        )

        request = tm11.ManageCollectionSubscriptionRequest(**request_params)
        response = self._execute_request(
            request, uri=uri,
            service_type=const.SVC_COLLECTION_MANAGEMENT)

        return to_subscription_response_entity(response, version=11)

    def get_subscription_status(self, collection_name,
                                subscription_id=None, uri=None):
        '''Get subscription status from TAXII Collection Management service.

        Sends a subscription request with action `STATUS`.
        If no ``subscription_id`` is provided, server will return the list
        of all available subscriptions for a collection with a name
        specified in ``collection_name``.

        if ``uri`` is not provided, client will try to discover services and
        find Collection Management Service among them.

        :param str collection_name: target collection name
        :param str subscription_id: subscription ID (optional)
        :param str uri: URI path to a specific Collection Management service

        :return: subscription information response
        :rtype: :py:class:`cabby.entities.SubscriptionResponse`

        :raises ValueError:
                if URI provided is invalid or schema is not supported
        :raises `cabby.exceptions.HTTPError`:
                if HTTP error happened
        :raises `cabby.exceptions.UnsuccessfulStatusError`:
                if Status Message received and status_type is not `SUCCESS`
        :raises `cabby.exceptions.ServiceNotFoundError`:
                if no Collection Management service found
        :raises `cabby.exceptions.AmbiguousServicesError`:
                more than one service with type specified
        :raises `cabby.exceptions.NoURIProvidedError`:
                no URI provided and client can't discover services
        '''

        return self.__subscription_status_request(
            const.ACT_STATUS, collection_name, subscription_id=subscription_id,
            uri=uri)

    def pause_subscription(self, collection_name, subscription_id, uri=None):
        '''Pause a subscription.

        Sends a subscription request with action `PAUSE`.  Subscription is
        identified by ``collection_name`` and ``subscription_id``.

        if ``uri`` is not provided, client will try to discover services and
        find Collection Management Service among them.

        :param str collection_name: target collection name
        :param str subscription_id: subscription ID
        :param str uri: URI path to a specific Collection Management service

        :return: subscription information response
        :rtype: :py:class:`cabby.entities.SubscriptionResponse`

        :raises ValueError:
                if URI provided is invalid or schema is not supported
        :raises `cabby.exceptions.HTTPError`:
                if HTTP error happened
        :raises `cabby.exceptions.UnsuccessfulStatusError`:
                if Status Message received and status_type is not `SUCCESS`
        :raises `cabby.exceptions.ServiceNotFoundError`:
                if no service found
        :raises `cabby.exceptions.AmbiguousServicesError`:
                more than one service with type specified
        :raises `cabby.exceptions.NoURIProvidedError`:
                no URI provided and client can't discover services
        '''

        return self.__subscription_status_request(
            const.ACT_PAUSE, collection_name, subscription_id=subscription_id,
            uri=uri)

    def resume_subscription(self, collection_name, subscription_id, uri=None):
        '''Resume a subscription.

        Sends a subscription request with action `RESUME`.
        Subscription is identified by ``collection_name``
        and ``subscription_id``.

        if ``uri`` is not provided, client will try to discover services and
        find Collection Management Service among them.

        :param str collection_name: target collection name
        :param str subscription_id: subscription ID
        :param str uri: URI path to a specific Collection Management service

        :return: subscription information response
        :rtype: :py:class:`cabby.entities.SubscriptionResponse`

        :raises ValueError:
                if URI provided is invalid or schema is not supported
        :raises `cabby.exceptions.HTTPError`:
                if HTTP error happened
        :raises `cabby.exceptions.UnsuccessfulStatusError`:
                if Status Message received and status_type is not `SUCCESS`
        :raises `cabby.exceptions.ServiceNotFoundError`:
                if no service found
        :raises `cabby.exceptions.AmbiguousServicesError`:
                more than one service with type specified
        :raises `cabby.exceptions.NoURIProvidedError`:
                no URI provided and client can't discover services
        '''

        return self.__subscription_status_request(
            const.ACT_RESUME, collection_name,
            subscription_id=subscription_id, uri=uri)

    def unsubscribe(self, collection_name, subscription_id, uri=None):
        '''Unsubscribe from a subscription.

        Sends a subscription request with action `UNSUBSCRIBE`. Subscription
        is identified by ``collection_name`` and ``subscription_id``.

        if ``uri`` is not provided, client will try to discover services and
        find Collection Management Service among them.

        :param str collection_name: target collection name
        :param str subscription_id: subscription ID
        :param str uri: URI path to a specific TAXII service

        :return: subscription information response
        :rtype: :py:class:`cabby.entities.SubscriptionResponse`

        :raises ValueError:
                if URI provided is invalid or schema is not supported
        :raises `cabby.exceptions.HTTPError`:
                if HTTP error happened
        :raises `cabby.exceptions.UnsuccessfulStatusError`:
                if Status Message received and status_type is not `SUCCESS`
        :raises `cabby.exceptions.ServiceNotFoundError`:
                if no service found
        :raises `cabby.exceptions.AmbiguousServicesError`:
                more than one service with type specified
        :raises `cabby.exceptions.NoURIProvidedError`:
                no URI provided and client can't discover services
        '''

        return self.__subscription_status_request(
            const.ACT_UNSUBSCRIBE, collection_name,
            subscription_id=subscription_id, uri=uri)

    def subscribe(self, collection_name, count_only=False, inbox_service=None,
                  content_bindings=None, uri=None):
        '''Create a subscription.

        Sends a subscription request with action `SUBSCRIBE`.

        if ``uri`` is not provided, client will try to discover services and
        find Collection Management Service among them.

        :param str collection_name: target collection name
        :param bool count_only: subscribe only to counts and not full content
        :param `cabby.entities.InboxService` inbox_service:
                Inbox Service that will accept content pushed by TAXII Server
                in the context of this subscription
        :param list content_bindings: a list of strings or
                :py:class:`cabby.entities.ContentBinding` entities
        :param str uri: URI path to a specific Collection Management service

        :return: subscription information response
        :rtype: :py:class:`cabby.entities.SubscriptionResponse`

        :raises ValueError:
                if URI provided is invalid or schema is not supported
        :raises `cabby.exceptions.HTTPError`:
                if HTTP error happened
        :raises `cabby.exceptions.UnsuccessfulStatusError`:
                if Status Message received and status_type is not `SUCCESS`
        :raises `cabby.exceptions.ServiceNotFoundError`:
                if no service found
        :raises `cabby.exceptions.AmbiguousServicesError`:
                more than one service with type specified
        :raises `cabby.exceptions.NoURIProvidedError`:
                no URI provided and client can't discover services
        '''

        response_type = const.RT_COUNT_ONLY if count_only else const.RT_FULL

        sparams = tm11.SubscriptionParameters(
            response_type=response_type,
            content_bindings=pack_content_bindings(content_bindings,
                                                   version=11)
        )
        rparams = dict(
            message_id=self._generate_id(),
            action=const.ACT_SUBSCRIBE,
            collection_name=collection_name,
            subscription_parameters=sparams,
        )

        if inbox_service:
            binding = (inbox_service.message_bindings[0]
                       if inbox_service.message_bindings else '')

            rparams['push_parameters'] = tm11.PushParameters(
                inbox_protocol=inbox_service.protocol,
                inbox_address=inbox_service.address,
                delivery_message_binding=binding
            )

        request = tm11.ManageCollectionSubscriptionRequest(**rparams)
        response = self._execute_request(
            request, uri=uri,
            service_type=const.SVC_COLLECTION_MANAGEMENT)

        return to_subscription_response_entity(response, version=11)

    def get_collections(self, uri=None):
        '''Get collections from Collection Management Service.

        if ``uri`` is not provided, client will try to discover services and
        find Collection Management Service among them.

        :param str uri: URI path to a specific Collection Management service

        :return: list of collections
        :rtype: list of :py:class:`cabby.entities.Collection`

        :raises ValueError:
                if URI provided is invalid or schema is not supported
        :raises `cabby.exceptions.HTTPError`:
                if HTTP error happened
        :raises `cabby.exceptions.UnsuccessfulStatusError`:
                if Status Message received and status_type is not `SUCCESS`
        :raises `cabby.exceptions.ServiceNotFoundError`:
                if no service found
        :raises `cabby.exceptions.AmbiguousServicesError`:
                more than one service with type specified
        :raises `cabby.exceptions.NoURIProvidedError`:
                no URI provided and client can't discover services
        '''

        request = tm11.CollectionInformationRequest(
            message_id=self._generate_id())

        response = self._execute_request(
            request, uri=uri,
            service_type=const.SVC_COLLECTION_MANAGEMENT)

        return to_collection_entities(response.collection_informations,
                                      version=11)

    def push(self, content, content_binding, collection_names=None,
             timestamp=None, uri=None):
        '''Push content into Inbox Service.

        if ``uri`` is not provided, client will try to discover
        services and find Inbox Service among them.

        :param str content: content to push
        :param content_binding: content binding for a content
        :type content_binding: string or
                               :py:class:`cabby.entities.ContentBinding`
        :param list collection_names:
                destination collection names
        :param datetime timestamp: timestamp label of the content block
                (current UTC time by default)
        :param str uri: URI path to a specific Inbox Service

        :raises ValueError:
                if URI provided is invalid or schema is not supported
        :raises `cabby.exceptions.HTTPError`:
                if HTTP error happened
        :raises `cabby.exceptions.UnsuccessfulStatusError`:
                if Status Message received and status_type is not `SUCCESS`
        :raises `cabby.exceptions.ServiceNotFoundError`:
                if no service found
        :raises `cabby.exceptions.AmbiguousServicesError`:
                more than one service with type specified
        :raises `cabby.exceptions.NoURIProvidedError`:
                no URI provided and client can't discover services
        '''

        content_block = tm11.ContentBlock(
            content=content,
            content_binding=pack_content_binding(content_binding, version=11),
            timestamp_label=timestamp or get_utc_now()
        )

        inbox_message = tm11.InboxMessage(message_id=self._generate_id(),
                                          content_blocks=[content_block])

        if collection_names:
            inbox_message.destination_collection_names.extend(collection_names)

        self._execute_request(inbox_message, uri=uri,
                              service_type=const.SVC_INBOX)

        self.log.debug("Content block successfully pushed")

    def _prepare_poll_request(self, collection_name, begin_date=None,
                              end_date=None, subscription_id=None,
                              inbox_service=None, content_bindings=None,
                              count_only=False):
        data = dict(
            message_id=self._generate_id(),
            collection_name=collection_name,
            exclusive_begin_timestamp_label=begin_date,
            inclusive_end_timestamp_label=end_date
        )

        if subscription_id:
            data['subscription_id'] = subscription_id
        else:
            _bindings = pack_content_bindings(content_bindings, version=11)
            poll_params = {'content_bindings': _bindings}

            if inbox_service:
                message_bindings = inbox_service.message_bindings[0] \
                    if inbox_service.message_bindings else []

                poll_params['delivery_parameters'] = tm11.DeliveryParameters(
                    inbox_protocol=inbox_service.protocol,
                    inbox_address=inbox_service.address,
                    delivery_message_binding=message_bindings
                )
                poll_params['allow_asynch'] = True

            if count_only:
                poll_params['response_type'] = const.RT_COUNT_ONLY
            else:
                poll_params['response_type'] = const.RT_FULL

            data['poll_parameters'] = \
                tm11.PollRequest.PollParameters(**poll_params)

        return tm11.PollRequest(**data)

    def get_content_count(self, collection_name, begin_date=None,
                          end_date=None, subscription_id=None,
                          inbox_service=None, content_bindings=None, uri=None):
        '''Get content blocks count for a query.

        if ``uri`` is not provided, client will try to discover services and
        find Polling Service among them.

        If ``subscription_id`` provided, arguments ``content_bindings`` and
        ``inbox_service`` are ignored.

        :param str collection_name: collection to poll
        :param datetime begin_date: ask only for content blocks created
               after `begin_date` (exclusive)
        :param datetime end_date: ask only for content blocks created
               before `end_date` (inclusive)
        :param str subsctiption_id: ID of the existing subscription
        :param `cabby.entities.InboxService` inbox_service:
               Inbox Service that will accept content pushed by TAXII Server
               in the context of this Poll Request
        :param list content_bindings: list of stings or
               :py:class:`cabby.entities.ContentBinding` objects
        :param str uri: URI path to a specific Inbox Service

        :raises ValueError:
                if URI provided is invalid or schema is not supported
        :raises `cabby.exceptions.HTTPError`:
                if HTTP error happened
        :raises `cabby.exceptions.UnsuccessfulStatusError`:
                if Status Message received and status_type is not `SUCCESS`
        :raises `cabby.exceptions.ServiceNotFoundError`:
                if no service found
        :raises `cabby.exceptions.AmbiguousServicesError`:
                more than one service with type specified
        :raises `cabby.exceptions.NoURIProvidedError`:
                no URI provided and client can't discover services

        :rtype: :py:class:`cabby.entities.ContentBlockCount`
        '''

        request = self._prepare_poll_request(
            collection_name,
            begin_date=begin_date,
            end_date=end_date,
            subscription_id=subscription_id,
            inbox_service=inbox_service,
            content_bindings=content_bindings,
            count_only=True
        )
        response = self._execute_request(
            request, uri=uri, service_type=const.SVC_POLL)

        for obj in response:
            if isinstance(obj, tm11.PollResponse):
                if obj.record_count:
                    return ContentBlockCount(
                        count=obj.record_count.record_count,
                        is_partial=obj.record_count.partial_count
                    )
                else:
                    return None

    def poll(self, collection_name, begin_date=None, end_date=None,
             subscription_id=None, inbox_service=None,
             content_bindings=None, uri=None):
        '''Poll content from Polling Service.

        if ``uri`` is not provided, client will try to discover services and
        find Polling Service among them.

        If ``subscription_id`` provided, arguments ``content_bindings`` and
        ``inbox_service`` are ignored.

        :param str collection_name: collection to poll
        :param datetime begin_date: ask only for content blocks created
               after `begin_date` (exclusive)
        :param datetime end_date: ask only for content blocks created
               before `end_date` (inclusive)
        :param str subsctiption_id: ID of the existing subscription
        :param `cabby.entities.InboxService` inbox_service:
               Inbox Service that will accept content pushed by TAXII Server
               in the context of this Poll Request
        :param list content_bindings: list of stings or
               :py:class:`cabby.entities.ContentBinding` objects
        :param str uri: URI path to a specific Inbox Service

        :raises ValueError:
                if URI provided is invalid or schema is not supported
        :raises `cabby.exceptions.HTTPError`:
                if HTTP error happened
        :raises `cabby.exceptions.UnsuccessfulStatusError`:
                if Status Message received and status_type is not `SUCCESS`
        :raises `cabby.exceptions.ServiceNotFoundError`:
                if no service found
        :raises `cabby.exceptions.AmbiguousServicesError`:
                more than one service with type specified
        :raises `cabby.exceptions.NoURIProvidedError`:
                no URI provided and client can't discover services
        '''

        request = self._prepare_poll_request(
            collection_name,
            begin_date=begin_date,
            end_date=end_date,
            subscription_id=subscription_id,
            inbox_service=inbox_service,
            content_bindings=content_bindings,
            count_only=False
        )

        stream = self._execute_request(request, uri=uri,
                                       service_type=const.SVC_POLL)
        response = None
        for obj in stream:
            if isinstance(obj, tm11.ContentBlock):
                yield to_content_block_entity(obj)
            else:
                response = obj
                break

        if response and response.more:
            part = response.result_part_number

            while True:
                part += 1
                has_data = False

                fulfilment_stream = self.fulfilment(
                    collection_name, response.result_id,
                    part_number=part, uri=uri)

                for block in fulfilment_stream:
                    has_data = True
                    yield block

                if not has_data:
                    break

    def fulfilment(self, collection_name, result_id, part_number=1, uri=None):
        '''Poll content from Polling Service as a part of fulfilment process.

        if ``uri`` is not provided, client will try to discover services and
        find Polling Service among them.

        :param str collection_name: collection to poll
        :param str result_id: existing polling Result ID
        :param int part_number: index number of a part from the result set
        :param str uri: URI path to a specific Inbox Service

        :raises ValueError:
                if URI provided is invalid or schema is not supported
        :raises `cabby.exceptions.HTTPError`:
                if HTTP error happened
        :raises `cabby.exceptions.UnsuccessfulStatusError`:
                if Status Message received and status_type is not `SUCCESS`
        :raises `cabby.exceptions.ServiceNotFoundError`:
                if no service found
        :raises `cabby.exceptions.AmbiguousServicesError`:
                more than one service with type specified
        :raises `cabby.exceptions.NoURIProvidedError`:
                no URI provided and client can't discover services
        '''

        request = tm11.PollFulfillmentRequest(
            message_id=self._generate_id(),
            collection_name=collection_name,
            result_id=result_id,
            result_part_number=part_number
        )

        stream = self._execute_request(request, uri=uri,
                                       service_type=const.SVC_POLL)

        for obj in stream:
            if isinstance(obj, tm11.ContentBlock):
                yield to_content_block_entity(obj)
